/*
Template functions lower_bound_, upper_bound_, equal_range and binary_search,
though implemented in the standard library, can be re-expressed in terms of
partition_point.

A specific type of iterators needs its specialized version of partition_point,
and can reuse unified definitions of lower/upper_bound etc.
*/

#ifndef PARTITION_HPP
#define PARTITION_HPP

#if defined(_MSC_VER)&&(_MSC_VER>=1200)
#pragma once
#endif


#include <boost/preprocessor/seq/enum.hpp>
#include <boost/preprocessor/seq/elem.hpp>
#include <boost/preprocessor/seq/transform.hpp>
#include <functional>
#include <utility>

/*
Expands to definitions of template functions partition_point, upper_bound,
lower_bound, equal_range and binary_search, specialized for a particular type
of iterators.

Usage:

BOOST_CUSTOM_PARTITION(seqParams, IterTmplForm, Begin, End, PredType, Pred) {
  implementation of partition_point for this kind of iterators
}

where

seqParams is a BOOST sequence of the form
    ((class)(A))((typename)(B))...
    Each element of seqParams is a sequence of 2 elements: the first is a
    keyword 'class' or 'typename', the second is a symbolic name of the
    parameter. All template parameters for the iterator type must be listed
    in this sequence. Example:
    ((class)(Value))((class)(Functor))((class)(Filter))

IterTmplForm is a syntactical form of the iterator template type where commas
    are replaced with BOOST_PP_COMMA(), and all template parameters are listed
    in seqParams; for example,
    my_iterator<Value BOOST_PP_COMMA() Functor BOOST_PP_COMMA() Filter>

Begin, End are names to be used as formal parameters of the custom function
   partition_point - begin and end of the iterator range.

PredType is a name to be used as a template parameter name in the definition
    of custom partition_point function (type of a unary predicate).

Pred is a name to be used for the third formal parameter of the custom function
    partition_point.
*/
#define BOOST_CUSTOM_PARTITION(_seq,_It,_begin,_end,_unaryPred,_pred)                   \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val, class BinPred>  \
  _It lower_bound(_It itBegin,_It itEnd,const Val &x,BinPred pred){               \
    return partition_point(itBegin,itEnd,!boost::bind(pred,_1,x));                \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val>                 \
  _It lower_bound(_It itBegin,_It itEnd,const Val &x){                            \
    return partition_point(itBegin,itEnd,!boost::bind(std::less<Val>(),_1,x));    \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val, class BinPred>  \
  _It upper_bound(_It itBegin,_It itEnd,const Val &x,BinPred pred){               \
  	return partition_point(itBegin,itEnd,boost::bind(pred,x,_1));                 \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val>                 \
  _It upper_bound(_It itBegin,_It itEnd,const Val &x){                            \
  	return partition_point(itBegin,itEnd,boost::bind(std::less<Val>(),x,_1));     \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val, class BinPred>  \
  std::pair<_It,_It>equal_range(_It itBegin,_It itEnd,const Val &x,BinPred pred){ \
    using namespace std;                                                          \
    itBegin=lower_bound(itBegin,itEnd,x,pred);                                    \
    return make_pair(itBegin,upper_bound(itBegin,itEnd,x,pred));                  \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val>                 \
  std::pair<_It,_It>equal_range(_It itBegin,_It itEnd,const Val &x){              \
    using namespace std;                                                          \
    itBegin=lower_bound(itBegin,itEnd,x);                                         \
    return make_pair(itBegin,upper_bound(itBegin,itEnd,x));                       \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val, class BinPred>  \
  bool binary_search(_It itBegin,_It itEnd,const Val &x,BinPred pred){            \
    using namespace std;                                                          \
    itBegin = lower_bound(itBegin,itEnd,x,pred);                                  \
    if(itBegin==itEnd) return false;                                              \
    return !(pred(x,*itBegin));                                                   \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class Val>                 \
  bool binary_search(_It itBegin,_It itEnd,const Val &x){                         \
    using namespace std;                                                          \
    itBegin = lower_bound(itBegin,itEnd,x);                                       \
    if(itBegin==itEnd) return false;                                              \
	return !(x < *itBegin);                                                       \
  }                                                                               \
  template<COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq),class _unaryPred>          \
    _It partition_point(_It _begin, _It _end, _unaryPred _pred)

#define SPACE_DELIMITED_PAIR(_s, _data, _elem)                                    \
  BOOST_PP_SEQ_ELEM(0, _elem) BOOST_PP_SEQ_ELEM(1, _elem)

#define COMMA_SEPARATED_SPACE_DELIMITED_PAIRS(_seq)                               \
  BOOST_PP_SEQ_ENUM(BOOST_PP_SEQ_TRANSFORM(SPACE_DELIMITED_PAIR,0,_seq))

#endif
