#ifndef PARTITION_POINT_HPP
#define PARTITION_POINT_HPP

#if defined(_MSC_VER)&&(_MSC_VER>=1200)
#pragma once
#endif

#include <boost/iterator/iterator_traits.hpp>

namespace boost {

/*
TEMPLATE PARAMETERS:

It - iterator type in some container;

UnaryPred - a type of unary predicate defined on the type of elements
referenced by iterators.

ARGUMENTS

itBegin, itEnd - two iterators that form a range;
pred - a unary predicate defined for values in the range.

PRECONDITIONS

[itBegin, itEnd) must be a valid range in some container: itBegin must be
    dereferenceable, itEnd must be achievable from itBegin by increment.

pred must be a partitioning of [itBegin, itEnd) range. For all iterators i, j
   from that range, if pred is false on (*i) and true on (*j), j must succeed i
   in the container. I.e. false cannot occur after true.

LEMMA

If there exist iterators i and j in the range [itBegin, itEnd), such that pred
    is false on (*i) and true on (*j), then there must be an iterator k, such
    that pred is false on (*p) for all iterators p that precede k and true on
    (*q) for all iterators q that succeed or are equal to k.
In other words, if the partitioning predicate changes its boolean value
	somewhere in the range, the value-changing point can be located.

RETURN VALUE

The leftmost iterator k in the range [itBegin, itEnd) such that pred is true
    on (*k), or itEnd if pred is false on all elements of the range.

REMARKS

This is the most general, universal version of partition_point. Works with any
    iterators but doesn't use any specific properties of data structures that
    could improve efficiency.
*/
template<class It, class UnaryPred> inline
It partition_point( It itBegin, It itEnd, UnaryPred pred ) {
#ifdef _DEBUG /* is pred a partitioning? False must not occur after true */
	It itPartitionPoint = std::find_if(itBegin, itEnd, pred);
	_ASSERT( std::find_if(itPartitionPoint, itEnd, !bind(pred, _1)) == itEnd );
#endif
	iterator_difference<It>::type nCount = std::distance(itBegin, itEnd);
	while( 0 < nCount ) {
		iterator_difference<It>::type nCount2 = nCount / 2;
		It itMid = itBegin;
		std::advance(itMid, nCount2);
		if( pred(*itMid) ) {
			nCount = nCount2;
		} else {
			itBegin = ++itMid;
			nCount -= nCount2 + 1;
		}
	}
#ifdef _DEBUG /* we have already found the partition point; compare result */
	_ASSERT( itBegin == itPartitionPoint );
#endif
	return itBegin;
}

}; /* namespace boost */

#endif /* PARTITION_POINT_HPP */
